"""
A module for working with schemas belonging to this package.
"""

# built-in
from typing import Any
from typing import Optional as _Optional

# third-party
from vcorelib.dict.codec import DictCodec as _DictCodec
from vcorelib.io import DEFAULT_INCLUDES_KEY
from vcorelib.schemas.base import SchemaMap as _SchemaMap
from vcorelib.schemas.json import JsonSchemaMap as _JsonSchemaMap

# internal
from ifgen import PKG_NAME


class IfgenDictCodec(_DictCodec):
    """
    A simple wrapper for package classes that want to implement DictCodec.
    """

    data: dict[str, Any]

    default_schemas: _Optional[_SchemaMap] = _JsonSchemaMap.from_package(
        PKG_NAME,
        includes_key=DEFAULT_INCLUDES_KEY,
    )



# =====================================
# generator=datazen
# version=3.2.3
# hash=91b95d05935cfffb260a5bf6ce0786f0
# =====================================
"""
ifgen - Package's default entry-point.
"""

# built-in
import sys

# internal
from ifgen.entry import main

if __name__ == "__main__":
    sys.exit(main(sys.argv))



# =====================================
# generator=datazen
# version=3.2.3
# hash=8042bfab9549cff58935b1dd444a1e8f
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "An interface generator for distributed computing."
PKG_NAME = "ifgen"
VERSION = "4.5.0"



"""
A module implementing an SVD-command configuration interface for the package.
"""

# third-party
from vcorelib.dict.codec import BasicDictCodec as _BasicDictCodec

# internal
from ifgen.schemas import IfgenDictCodec


class SvdConfig(IfgenDictCodec, _BasicDictCodec):
    """The top-level configuration object for the package."""



"""
Utility interfaces related to struct code generation.
"""

# internal
from ifgen.generation.interface import GenerateTask


def struct_dependencies(task: GenerateTask) -> set[str]:
    """Generates string type names for dependencies."""

    unique = set()

    for config in task.instance["fields"]:
        if "type" in config:
            unique.add(config["type"])

        # Add includes for bit-fields.
        for bit_field in config.get("fields", []):
            if "type" in bit_field:
                unique.add(bit_field["type"])

        # Add includes for alternates.
        for alternate in config.get("alternates", []):
            for alternate_bit_field in alternate.get("fields", []):
                if "type" in alternate_bit_field:
                    unique.add(alternate_bit_field["type"])

    return unique



"""
A module implementing Python struct-generation interfaces.
"""

# built-in
from typing import Any

# third-party
from runtimepy.codec.system import TypeSystem
from vcorelib.io import IndentedFileWriter
from vcorelib.names import to_snake

# internal
from ifgen.enum.python import strip_t_suffix
from ifgen.generation.interface import GenerateTask
from ifgen.generation.python import (
    PythonImports,
    python_class,
    python_function,
    python_imports,
)
from ifgen.struct.util import struct_dependencies


def cpp_ns_final(data: str) -> str:
    """Get only the final string in a C++ style namespace string."""
    return data.split("::")[-1]


def python_enums_structs(task: GenerateTask) -> tuple[list[str], list[str]]:
    """Get enum and struct dependency information for this task."""

    enums = []
    structs = []

    types = task.env.types
    for dep in struct_dependencies(task):
        dep = strip_t_suffix(dep)

        if dep in types.primitives:
            continue

        if types.is_enum(dep):
            enums.append(dep)
            continue

        if types.is_custom(dep):
            structs.append(dep)

    return sorted(enums), sorted(structs)


def python_dependencies(
    enums: list[str],
    structs: list[str],
    enum_prefix: str = "..enums",
    struct_prefix: str = "",
) -> PythonImports:
    """
    Get the names of external (to the module being generated) dependencies
    by type.
    """

    deps: PythonImports = {}

    for enum in enums:
        enum = cpp_ns_final(enum)
        deps[f"{enum_prefix}.{to_snake(enum)}"] = [enum]
    for struct in structs:
        struct = cpp_ns_final(struct)
        deps[f"{struct_prefix}.{to_snake(struct)}"] = [struct]

    return deps


def python_struct_field(
    writer: IndentedFileWriter,
    types: TypeSystem,
    field: dict[str, Any],
    structs: list[str],
    enums: list[str],
) -> None:
    """Write lines for a struct field."""

    writer.write("protocol.add_field(")

    kind = strip_t_suffix(field["type"])

    with writer.indented():
        line = f"\"{field['name']}\","
        if field.get("description"):
            line += f"  # {field['description']}"
        writer.write(line)

        if kind in types.primitives:
            writer.write(f'kind="{kind}",')

        elif kind in structs:
            writer.write("serializable=" f"{cpp_ns_final(kind)}.instance(),")

        elif kind in enums:
            writer.write(f'enum="{cpp_ns_final(kind)}",')

        if "array_length" in field:
            writer.write(f"array_length={field['array_length']},")

    writer.write(")")


def python_struct_fields(
    task: GenerateTask,
    writer: IndentedFileWriter,
    structs: list[str],
    enums: list[str],
) -> None:
    """Write protocol initialization code."""

    types = task.env.types

    # 'alternates' field not handled yet.

    for field in task.instance["fields"]:
        if "fields" in field:
            writer.write(
                "with protocol.add_bit_fields"
                f'("{field["name"]}", '
                f'kind="{strip_t_suffix(field["type"])}") as fields:'
            )
            with writer.indented():
                for bit_field in field["fields"]:
                    line = ""

                    if bit_field["width"] == 1:
                        line = "fields.flag("
                        line += f"\"{bit_field['name']}\", "
                    else:
                        line = "fields.field("
                        line += (
                            f"\"{bit_field['name']}\", "
                            f"{bit_field['width']}, "
                        )

                    if bit_field.get("description"):
                        line += f"description=\"{bit_field['description']}\", "

                    kind = bit_field.get("type")
                    if kind and kind in enums:
                        line += f'enum="{cpp_ns_final(kind)}", '

                    line += f"index={bit_field['index']}"

                    writer.write(line + ")")

        else:
            python_struct_field(writer, types, field, structs, enums)


def python_struct_header(
    task: GenerateTask, writer: IndentedFileWriter
) -> None:
    """Create a Python module for a struct."""

    enums, structs = python_enums_structs(task)

    python_imports(
        writer,
        third_party={
            "runtimepy.codec.protocol": ["Protocol", "ProtocolFactory"],
            "runtimepy.primitives.byte_order": ["STD_ENDIAN", "enum_registry"],
        },
        internal=python_dependencies(enums, structs),
    )

    proto = task.protocol()

    id_underlying = strip_t_suffix(
        task.env.config.data["struct"]["id_underlying"]
    )

    with python_class(
        writer,
        task.name,
        task.resolve_description() or "No description.",
        parents=["ProtocolFactory"],
        final_empty=0,
    ):
        # Create protocol instance.
        writer.write("protocol = Protocol(")
        with writer.indented():
            writer.write(
                f"enum_registry({', '.join(cpp_ns_final(x) for x in enums)}),"
            )
            writer.write(f"identifier={proto.id},")
            writer.write("identifier_primitive=" f'"{id_underlying}",')
            writer.write(
                "byte_order=STD_ENDIAN"
                f"[\"{task.instance['default_endianness']}\"],"
            )
        writer.write(")")

        writer.empty()

        with python_function(
            writer,
            "initialize",
            "Initialize this protocol.",
            params="cls, protocol: Protocol",
            decorators=["classmethod"],
            final_empty=0,
        ):
            python_struct_fields(task, writer, structs, enums)



"""
A module implementing enumerations used by this package.
"""

# built-in
from enum import StrEnum

# third-party
from vcorelib.names import to_snake

# internal
from ifgen import PKG_NAME


class Generator(StrEnum):
    """An enumeration declaring all valid kinds of generators."""

    STRUCTS = "structs"
    ENUMS = "enums"
    IFGEN = PKG_NAME
    CUSTOM = "custom"


class Language(StrEnum):
    """An enumeration declaring output generation variants."""

    CPP = "CPP"
    PYTHON = "Python"

    @property
    def source_suffix(self) -> str:
        """Get a source-file suffix for this language."""
        return "cc" if self is Language.CPP else "py"

    @property
    def header_suffix(self) -> str:
        """Get a header-file suffix for this language."""
        return "h" if self is Language.CPP else "py"

    @property
    def slug(self) -> str:
        """Get a slug string."""
        return to_snake(self.name)

    @property
    def cfg_dir_name(self) -> str:
        """
        Get the configuration key for this language's output configuration.
        """
        return f"{self.slug}_dir"



"""
Utilities shared between struct methods.
"""

# built-in
from contextlib import ExitStack

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask


def swap_fields(
    task: GenerateTask, writer: IndentedFileWriter, elem_prefix: str = ""
) -> None:
    """Generate code for swapping struct fields."""

    for field in task.instance["fields"]:
        kind = field["type"]

        if field["padding"] or task.env.size(kind) == 1:
            continue

        name = field["name"]

        with ExitStack() as stack:
            is_array = "array_length" in field
            if is_array:
                array_cmp = task.cpp_namespace(f"{name}_length")
                writer.write(f"for (std::size_t i = 0; i < {array_cmp}; i++)")
                stack.enter_context(writer.scope())

            elem = elem_prefix + name
            if is_array:
                elem += "[i]"

            if task.env.is_struct(kind):
                writer.write(f"{elem}.endian<endianness>();")
            elif task.env.is_enum(kind):
                writer.write(f"{elem} = handle_endian<endianness>({elem});")
            else:
                writer.write(f"{elem} = handle_endian<endianness>({elem});")



"""
A module implementing base interfaces for processing a group of peripherals.
"""

# built-in
from dataclasses import dataclass
from typing import Iterator, Optional

# internal
from ifgen.svd.model.peripheral import Peripheral


@dataclass
class PeripheralGroup:
    """A container for peripherals that have the same register layout."""

    root: Peripheral
    derivatives: list[Peripheral]

    @property
    def peripherals(self) -> Iterator[Peripheral]:
        """Get all peripheral instances."""
        yield self.root
        yield from self.derivatives


PRUNE_STRUCTS = False


def get_derived(
    peripheral: Peripheral, peripherals: list[Peripheral]
) -> Optional[Peripheral]:
    """Determine if this peripheral is derived from any other peripheral."""

    result = None

    if peripheral.derived:
        result = peripheral.derived_elem

    # Check if this peripheral is equivalent to some other peripheral.
    elif PRUNE_STRUCTS:
        for other in peripherals:
            # Always return None if you get far enough to see yourself in the
            # list. That way this peripheral becomes the effective 'root'.
            if other is peripheral:
                break

            if not other.is_alternate() and not other.derived:
                if other == peripheral:
                    result = other
                    break

    return result


def peripheral_groups(
    peripherals: dict[str, Peripheral],
) -> dict[str, PeripheralGroup]:
    """Organize peripherals into groups."""

    result: dict[str, PeripheralGroup] = {}

    peripherals_list = list(peripherals[x] for x in sorted(peripherals))
    for peripheral in peripherals_list:
        name = peripheral.base_name()

        derived = get_derived(peripheral, peripherals_list)
        if derived is not None:
            name = derived.base_name()

        if name not in result:
            # Validate this later.
            result[name] = PeripheralGroup(None, [])  # type: ignore

        group = result[name]

        if group.root is None:
            group.root = derived if derived is not None else peripheral
        else:
            assert derived is not None
            result[derived.base_name()].derivatives.append(peripheral)

    # Validate groups.
    for name, group in result.items():
        assert group.root is not None, (name, group)

    return result



"""
A module implementing interfaces for generating bit-field methods.
"""

# built-in
from typing import Any

# third-party
from vcorelib.io.file_writer import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask


def handle_description(
    writer: IndentedFileWriter, field: dict[str, Any]
) -> None:
    """Handle writing an instance's description."""

    if field.get("description"):
        writer.empty()
        writer.write(field["description"])


def set_bit_method(
    task: GenerateTask,
    name: str,
    field: dict[str, Any],
    writer: IndentedFileWriter,
    method_slug: str,
) -> None:
    """Generate a 'set' method for a bit-field."""

    method = task.cpp_namespace(f"set_{method_slug}(){task.method_suffix()}")
    writer.empty()

    with writer.javadoc():
        writer.write(f"Set {name}'s {field['name']} bit.")
        handle_description(writer, field)

    writer.write("inline void " + method)

    with writer.scope():
        stmt = f"1u << {field['index']}u"
        writer.write(f"{name} |= {stmt};")


def clear_bit_method(
    task: GenerateTask,
    name: str,
    field: dict[str, Any],
    writer: IndentedFileWriter,
    method_slug: str,
) -> None:
    """Generate a 'clear' method for a bit-field."""

    method = task.cpp_namespace(f"clear_{method_slug}(){task.method_suffix()}")
    writer.empty()

    with writer.javadoc():
        writer.write(f"Clear {name}'s {field['name']} bit.")
        handle_description(writer, field)

    writer.write("inline void " + method)

    with writer.scope():
        stmt = f"~(1u << {field['index']}u)"
        writer.write(f"{name} &= {stmt};")


def bit_field_toggle_method(
    task: GenerateTask,
    name: str,
    field: dict[str, Any],
    writer: IndentedFileWriter,
    method_slug: str,
) -> None:
    """Generate a 'toggle' method for a bit-field."""

    set_bit_method(task, name, field, writer, method_slug)
    clear_bit_method(task, name, field, writer, method_slug)

    method = task.cpp_namespace(
        f"toggle_{method_slug}(){task.method_suffix()}"
    )
    writer.empty()

    with writer.javadoc():
        writer.write(f"Toggle {name}'s {field['name']} bit.")
        handle_description(writer, field)

    writer.write("inline void " + method)

    with writer.scope():
        stmt = f"1u << {field['index']}u"
        writer.write(f"{name} ^= {stmt};")



"""
A module implementing a data model for ARM CMSIS-SVD 'peripheral' data.
"""

# built-in
from dataclasses import dataclass
from typing import Iterable, Optional, Union
from xml.etree import ElementTree

# internal
from ifgen.svd.model.address_block import AddressBlock
from ifgen.svd.model.derived import DerivedMixin
from ifgen.svd.model.device import ARRAY_PROPERTIES, REGISTER_PROPERTIES
from ifgen.svd.model.field import FieldMap
from ifgen.svd.model.interrupt import Interrupt
from ifgen.svd.string import StringKeyVal

RegisterData = list[Union["Register", "Cluster"]]


@dataclass
class Cluster(DerivedMixin):
    """A container for cluster information."""

    derived_from: Optional["Cluster"]
    children: RegisterData
    peripheral: "Peripheral"

    def __eq__(self, other) -> bool:
        """Determine if two clusers are equivalent."""

        return isinstance(other, Cluster) and all(
            x == y for x, y in zip(self.children, other.children)
        )

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        return (
            ARRAY_PROPERTIES
            + [
                StringKeyVal("name", True),
                StringKeyVal("description", False),
                StringKeyVal("alternateCluster", False),
                StringKeyVal("headerStructName", False),
                StringKeyVal("addressOffset", True),
            ]
            + REGISTER_PROPERTIES
        )


def fields_equal(left: Optional[FieldMap], right: Optional[FieldMap]) -> bool:
    """Determine if two field maps are equivalent."""

    result = left is None and right is None

    if left is not None and right is not None and len(left) == len(right):
        for lkey, lvalue in left.items():
            result = lkey in right and lvalue == right[lkey]
            if not result:
                break

    return result


@dataclass
class Register(DerivedMixin):
    """A container for register information."""

    derived_from: Optional["Register"]
    fields: Optional[FieldMap]
    peripheral: "Peripheral"

    def __eq__(self, other) -> bool:
        """Determine if two registers are equivalent."""

        return isinstance(other, Register) and fields_equal(
            self.fields, other.fields
        )

    @property
    def bits(self) -> int:
        """Get the size of this register in bits."""
        result = self.raw_data.get("size", self.peripheral.bits)
        assert result is not None
        return int(result)

    @property
    def size(self) -> int:
        """Get the size of this register in bytes."""
        return self.bits // 8

    @property
    def alternate_group(self) -> Optional[str]:
        """Get this register's possible alternate group."""
        return self.raw_data.get("alternateGroup")

    @property
    def access(self) -> str:
        """Get the access setting for this register."""

        access = self.raw_data.get("access", self.peripheral.access)

        read = False
        write = False

        if access is None and self.fields is not None:
            for field in self.fields.values():
                if "access" in field.raw_data:
                    read |= "read" in field.raw_data["access"]
                    write |= "write" in field.raw_data["access"]

        if read and not write:
            return "read-only"
        if write and not read:
            return "write-only"
        return "read-write"

    @property
    def c_type(self) -> str:
        """Get the C type for this register."""
        return f"uint{self.bits}_t"

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        return (
            ARRAY_PROPERTIES
            + [
                StringKeyVal("name", True),
                StringKeyVal("displayName", False),
                StringKeyVal("description", False),
                StringKeyVal("alternateGroup", False),
                StringKeyVal("alternateRegister", False),
                StringKeyVal("addressOffset", True),
            ]
            + REGISTER_PROPERTIES
            + [
                # is enum
                StringKeyVal("dataType", False),
                # is enum
                StringKeyVal("modifiedWriteValues", False),
                # is enum
                StringKeyVal("readAction", False),
            ]
        )


def register_groups(registers: RegisterData) -> dict[str, list[Register]]:
    """Get groups of registers."""

    result: dict[str, list[Register]] = {}

    for item in registers:
        if isinstance(item, Register):
            alt = item.alternate_group
            if alt:
                if alt not in result:
                    result[alt] = []
                result[alt].append(item)

    return result


@dataclass
class Peripheral(DerivedMixin):
    """A container for peripheral information."""

    derived_from: Optional["Peripheral"]

    # Currently treated as metadata.
    interrupts: list[Interrupt]
    address_blocks: list[AddressBlock]

    registers: RegisterData

    def register_groups(self) -> dict[str, list[Register]]:
        """Get register groups."""
        return register_groups(self.registers)

    def __eq__(self, other) -> bool:
        """Determine if two peripherals are equivalent."""

        return isinstance(other, Peripheral) and all(
            x == y for x, y in zip(self.registers, other.registers)
        )

    @property
    def bits(self) -> Optional[int]:
        """Get size for this peripheral in bits."""

        result = self.derived_elem.raw_data.get("size")
        return int(result) if result is not None else None

    @property
    def access(self) -> Optional[str]:
        """Get the possible 'access' field default."""
        return self.derived_elem.raw_data.get("access")

    def base_name(self, lower: bool = True) -> str:
        """Get the base peripheral name."""

        result = self.name
        result = result.lower() if lower else result
        return result

    def handle_address_block(self, address_block: ElementTree.Element) -> None:
        """Handle an 'address_block' element."""
        self.address_blocks.append(AddressBlock.create(address_block))

    def handle_interrupt(self, interrupt: ElementTree.Element) -> None:
        """Handle an 'interrupt' element."""
        self.interrupts.append(Interrupt.create(interrupt))

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        return (
            ARRAY_PROPERTIES
            + [
                StringKeyVal("name", True),
                StringKeyVal("version", False),
                StringKeyVal("description", False),
                StringKeyVal("alternatePeripheral", False),
                StringKeyVal("groupName", False),
                StringKeyVal("prependToName", False),
                StringKeyVal("appendToName", False),
                StringKeyVal("headerStructName", False),
                StringKeyVal("disableCondition", False),
                StringKeyVal("baseAddress", True),
            ]
            + REGISTER_PROPERTIES
        )



"""
A module implementing interfaces for struct-file generation.
"""

# built-in
from typing import Iterable

# internal
from ifgen import PKG_NAME
from ifgen.generation.interface import GenerateTask
from ifgen.struct.header import struct_header
from ifgen.struct.test import create_struct_test
from ifgen.struct.util import struct_dependencies

__all__ = ["create_struct", "create_struct_test"]
FieldConfig = dict[str, int | str]


def header_for_type(name: str, task: GenerateTask) -> str:
    """Determine the header file to import for a given type."""

    candidate = task.custom_include(name)
    if candidate:
        return f'"{candidate}"'

    return ""


def struct_includes(task: GenerateTask) -> Iterable[str]:
    """Determine headers that need to be included for a given struct."""

    result = [header_for_type(x, task) for x in struct_dependencies(task)]
    result.append(f'"../{PKG_NAME}/common.h"')
    return result


def create_struct(task: GenerateTask) -> None:
    """Create a header file based on a struct definition."""

    with task.boilerplate(
        includes=struct_includes(task),
        json=task.instance.get("json", False),
        parent_depth=2,
    ) as writer:
        struct_header(task, writer)



"""
An entry-point for the 'svd' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace
from logging import getLogger
from pathlib import Path

# third-party
from vcorelib.args import CommandFunction as _CommandFunction
from vcorelib.paths import find_file

# internal
from ifgen import PKG_NAME
from ifgen.config.svd import SvdConfig
from ifgen.svd import register_processors
from ifgen.svd.group import base, enums
from ifgen.svd.task import SvdProcessingTask


def svd_cmd(args: _Namespace) -> int:
    """Execute the svd command."""

    register_processors()
    logger = getLogger(__name__)

    path = find_file(
        args.svd_file, package=PKG_NAME, logger=logger, include_cwd=True
    )
    assert path is not None, args.svd_file

    config = SvdConfig.decode(
        find_file(args.config, logger=logger, include_cwd=True)
    )

    # Only enable certain pruning strategies for certain processors.
    enable_pruning = path.with_suffix("").name in config.data.setdefault(
        "enable_pruning", []
    )
    enums.PRUNE_ENUMS = enable_pruning
    base.PRUNE_STRUCTS = enable_pruning

    SvdProcessingTask.svd(path, args.min_enum_width).generate_configs(
        args.output, config
    )

    return 0


DEFAULT_MIN_ENUM_WIDTH = 2


def add_svd_cmd(parser: _ArgumentParser) -> _CommandFunction:
    """Add svd-command arguments to its parser."""

    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=f"{PKG_NAME}-out",
        help=(
            "output directory for configuration "
            "files (default '%(default)s')"
        ),
    )

    parser.add_argument(
        "-c",
        "--config",
        default=f"package://{PKG_NAME}/svd.yaml",
        help="configuration rules to use (default: '%(default)s')",
    )

    parser.add_argument(
        "-m",
        "--min-enum-width",
        type=int,
        default=DEFAULT_MIN_ENUM_WIDTH,
        help=(
            "minimum number of enumeration elements to warrant "
            "generating an enumeration definition (default: %(default)s)"
        ),
    )

    parser.add_argument(
        "svd_file", type=str, help="path/uri to a CMSIS-SVD file"
    )

    return svd_cmd



"""
A module for generating shared headers and sources.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.generation.test import unit_test_boilerplate


def create_common_test(task: GenerateTask) -> None:
    """Create a unit test for the enum string-conversion methods."""

    if task.is_python:
        return

    with unit_test_boilerplate(task, declare_namespace=True) as writer:
        writer.cpp_comment("TODO.")


def endianness_single(writer: IndentedFileWriter) -> None:
    """Add methods for byte-sized primitives."""

    writer.c_comment("No action for byte-sized primitives.")
    writer.write(
        "template <std::endian endianness, byte_size T> "
        "inline void handle_endian_p(T *)"
    )
    with writer.scope():
        pass

    writer.write(
        "template <std::endian endianness, byte_size T> "
        "inline T handle_endian(T elem)"
    )
    with writer.scope():
        writer.write("return elem;")


def endianness_native(writer: IndentedFileWriter) -> None:
    """Add methods for native endianness (no swap)."""

    writer.c_comment("No action if endianness is native.")
    writer.write("template <std::endian endianness, std::integral T>")
    writer.write("inline void handle_endian_p(T *)")
    with writer.indented():
        writer.write(
            "requires(not byte_size<T>) && "
            "(endianness == std::endian::native)"
        )
    with writer.scope():
        pass

    writer.write("template <std::endian endianness, std::integral T>")
    writer.write("inline T handle_endian(T elem)")
    with writer.indented():
        writer.write(
            "requires(not byte_size<T>) && "
            "(endianness == std::endian::native)"
        )
    with writer.scope():
        writer.write("return elem;")


def endianness_integral(writer: IndentedFileWriter) -> None:
    """Add methods for integral types that require swapping."""

    writer.c_comment("Swap any integral type.")

    writer.write("template <std::endian endianness, std::integral T>")
    writer.write("inline T handle_endian(T elem)")
    with writer.indented():
        writer.write(
            "requires(not byte_size<T>) && (endianness != std::endian::native)"
        )
    with writer.scope():
        writer.write("return std::byteswap(elem);")

    writer.write("template <std::endian endianness, std::integral T>")
    writer.write("inline void handle_endian_p(T *elem)")
    with writer.indented():
        writer.write(
            "requires(not byte_size<T>) && (endianness != std::endian::native)"
        )
    with writer.scope():
        writer.write("*elem = std::byteswap(*elem);")


def endianness_enum(writer: IndentedFileWriter) -> None:
    """Add methods for enum types that require swapping."""

    writer.c_comment("Handler for enum class types.")
    writer.write("template <std::endian endianness, typename T>")
    writer.write("inline void handle_endian_p(T *elem)")
    with writer.indented():
        writer.write("requires(std::is_enum_v<T>)")
    with writer.scope():
        writer.write("using underlying = std::underlying_type_t<T>;")
        writer.write("handle_endian_p<endianness>(")
        with writer.indented():
            writer.write("reinterpret_cast<underlying *>(elem));")

    writer.write("template <std::endian endianness, typename T>")
    writer.write("inline T handle_endian(T elem)")
    with writer.indented():
        writer.write("requires(std::is_enum_v<T>)")
    with writer.scope():
        writer.write("return static_cast<T>(handle_endian<endianness>(")
        with writer.indented():
            writer.write("std::to_underlying(elem)));")


def endianness_float(writer: IndentedFileWriter) -> None:
    """Add methods for floating-point types that require swapping."""

    for width in ["32", "64"]:
        writer.empty()
        writer.c_comment(f"Handler for {width}-bit float.")
        writer.write(
            "template <std::endian endianness, std::floating_point T>"
        )
        writer.write("inline void handle_endian_p(T *elem)")

        prim = f"uint{width}_t"

        with writer.indented():
            writer.write(f"requires(sizeof(T) == sizeof({prim}))")
        with writer.scope():
            writer.write(
                f"handle_endian_p<endianness>"
                f"(reinterpret_cast<{prim} *>(elem));"
            )

        writer.write(
            "template <std::endian endianness, std::floating_point T>"
        )
        writer.write("inline T handle_endian(T elem)")
        with writer.indented():
            writer.write(f"requires(sizeof(T) == sizeof({prim}))")
        with writer.scope():
            writer.write("return std::bit_cast<T>(")
            with writer.indented():
                writer.write(
                    "handle_endian<endianness>"
                    f"(std::bit_cast<{prim}>(elem)));"
                )


def common_endianness(writer: IndentedFileWriter, task: GenerateTask) -> None:
    """Write endianness-related content."""

    writer.c_comment("Enforce that this isn't a mixed-endian system.")
    writer.write("static_assert(std::endian::native == std::endian::big or")
    writer.write("              std::endian::native == std::endian::little);")

    data = task.env.config.data
    endian = data["struct"]["default_endianness"]

    with writer.padding():
        writer.c_comment("Default endianness configured.")
        writer.write(
            f"static constexpr auto default_endian = std::endian::{endian};"
        )

    writer.c_comment("Detect primitives that don't need byte swapping.")
    writer.write("template <typename T>")
    writer.write("concept byte_size = sizeof(T) == 1;")

    with writer.padding():
        endianness_single(writer)

    endianness_native(writer)

    writer.empty()

    endianness_integral(writer)

    endianness_float(writer)

    writer.empty()

    endianness_enum(writer)


def create_common(task: GenerateTask) -> None:
    """Create a unit test for the enum string-conversion methods."""

    if task.is_python:
        return

    streams = task.stream_implementation

    includes = [
        "<bit>",
        "<concepts>",
        "<cstdint>",
        "<span>" if not streams else "<spanstream>",
        "<utility>",
    ]

    # probably get rid of everything besides the spanstream
    if streams:
        includes.extend(["<streambuf>", "<istream>", "<ostream>"])

    with task.boilerplate(includes=includes) as writer:
        common_endianness(writer, task)

        writer.empty()
        writer.c_comment("Configured primitives for identifiers.")
        data = task.env.config.data
        writer.write(f"using struct_id_t = {data['struct']['id_underlying']};")
        writer.write(f"using enum_id_t = {data['enum']['id_underlying']};")

        with writer.padding():
            writer.c_comment("Create useful aliases for bytes.")
            writer.write("template <std::size_t Extent = std::dynamic_extent>")
            writer.write("using byte_span = std::span<std::byte, Extent>;")
            writer.write(
                (
                    "template <std::size_t size> using byte_array = "
                    "std::array<std::byte, size>;"
                )
            )

        if streams:
            writer.c_comment("Abstract byte-stream interfaces.")
            writer.write("using byte_istream = std::basic_istream<std::byte>;")
            writer.write("using byte_ostream = std::basic_ostream<std::byte>;")

            writer.empty()
            writer.c_comment(
                "Concrete byte-stream interfaces (based on span)."
            )
            writer.write("using byte_spanbuf = std::basic_spanbuf<std::byte>;")
            writer.write(
                "using byte_spanstream = std::basic_spanstream<std::byte>;"
            )

        writer.empty()
        writer.c_comment("Constraint for generated structs.")
        writer.write("template <typename T>")
        writer.write("concept ifgen_struct = requires")
        with writer.scope(suffix=";"):
            writer.write("std::is_integral_v<decltype(T::id)>;")
            writer.write("std::is_same_v<decltype(T::size), std::size_t>;")



"""
A module implementing unit-testing related generation utilities.
"""

# built-in
from contextlib import ExitStack, contextmanager
from os import linesep
from typing import Iterator

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask


def unit_test_method_name(name: str, task: GenerateTask) -> str:
    """Get the name of a unit test."""
    return f"test_{task.name}_{name}"


@contextmanager
def unit_test_method(
    name: str, task: GenerateTask, writer: IndentedFileWriter
) -> Iterator[None]:
    """Generate unit-test method boilerplate."""

    writer.write("template <std::endian endianness>")
    writer.write(f"void {unit_test_method_name(name, task)}(void)")
    with writer.scope():
        nspace = task.env.types.root_namespace

        project_wide = nspace.namespace(track=False)
        writer.write(f"using namespace {project_wide};")

        with nspace.pushed(*task.instance.get("namespace", [])):
            curr = nspace.namespace(track=False)
            if curr != project_wide:
                writer.write(f"using namespace {curr};")

        writer.empty()
        yield


@contextmanager
def unit_test_main(
    task: GenerateTask, writer: IndentedFileWriter, description: bool = True
) -> Iterator[None]:
    """A method for generating main-function boilerplate for unit tests."""

    if description:
        with writer.javadoc():
            writer.write(f"A unit test for {task.generator} {task.name}.")
            writer.empty()
            writer.write("\\return 0 on success.")

    writer.write("int main(void)")
    with writer.scope():
        writer.write(f"using namespace {task.namespace()};")
        writer.empty()
        yield
        writer.write("return 0;")


@contextmanager
def unit_test_boilerplate(
    task: GenerateTask,
    includes: list[str] = None,
    main: bool = True,
    declare_namespace: bool = False,
) -> Iterator[IndentedFileWriter]:
    """Handle standard unit-test boilerplate."""

    include = task.env.rel_include(task.name, task.generator, task.language)

    if includes is None:
        includes = []

    linesep.join([f"A unit test for {task.generator} {task.name}."])

    with ExitStack() as stack:
        writer = stack.enter_context(
            task.boilerplate(
                includes=[
                    "<cassert>",
                    "<cstring>",
                    "<iostream>",
                    f'"{include}"',
                ]
                + includes,
                is_test=True,
                use_namespace=False,
            )
        )

        if declare_namespace:
            writer.write(f"namespace {task.namespace()}")
            with writer.scope(suffix=";"):
                writer.c_comment(
                    (
                        "Declared to ensure this namespace "
                        "has been declared in general."
                    )
                )
            writer.empty()

        if main:
            stack.enter_context(unit_test_main(task, writer))

        yield writer



"""
A module implementing interfaces for generating struct method code.
"""

# built-in
from typing import Any

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.generation.json import to_json_method
from ifgen.struct.methods.common import swap_fields


def protocol_json(task: GenerateTask) -> dict[str, Any]:
    """Get JSON data for this struct task."""

    protocol = task.protocol()

    # use something better for this

    return protocol.export_json()


def span_method(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Generate a span method."""

    with writer.javadoc():
        writer.write(("Get this instance as a byte span."))

    span_type = task.cpp_namespace("Span")
    method = task.cpp_namespace("span")

    writer.write(f"inline {span_type} {method}()")

    with writer.scope():
        writer.write("return Span(*raw());")


def struct_buffer_method(
    task: GenerateTask,
    writer: IndentedFileWriter,
    read_only: bool,
) -> None:
    """Generate a method for raw buffer access."""

    with writer.javadoc():
        writer.write(
            (
                "Get this instance as a "
                f"{'read-only ' if read_only else ''}"
                "fixed-size byte array."
            )
        )

    buff_type = task.cpp_namespace("Buffer")

    if read_only:
        buff_type = "const " + buff_type

    # Returns a pointer.
    method = task.cpp_namespace(
        "raw()" if not read_only else "raw_ro()", prefix="*"
    )
    writer.write(
        f"inline {buff_type} {method}" + (" const" if read_only else "")
    )

    with writer.scope():
        writer.write(
            "return reinterpret_cast"
            f"<{'const ' if read_only else ''}Buffer *>(this);"
        )


def endian_str(task: GenerateTask) -> str:
    """Get the appropriate default endian argument for this struct."""

    return (
        f"std::endian::{task.instance['default_endianness']}"
        if task.instance["default_endianness"]
        != task.env.config.data["struct"]["default_endianness"]
        else "default_endian"
    )


def handle_endian(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Write a struct method for byte swapping."""

    with writer.javadoc():
        writer.write("Handle swapping bytes for endian conversion (native).")
        writer.empty()
        writer.write(
            task.command(
                "tparam",
                "endianness Byte order for encoding elements.",
            )
        )

    writer.write(f"template <std::endian endianness = {endian_str(task)}>")
    writer.write("inline void endian(void)")
    with writer.indented():
        writer.write("requires(endianness == std::endian::native)")
    with writer.scope():
        pass

    writer.empty()

    with writer.javadoc():
        writer.write(
            "Handle swapping bytes for endian conversion (swap required)."
        )
        writer.empty()
        writer.write(
            task.command(
                "tparam",
                "endianness Byte order for encoding elements.",
            )
        )

    writer.write(f"template <std::endian endianness = {endian_str(task)}>")
    writer.write("inline void endian(void)")
    with writer.indented():
        writer.write("requires(endianness != std::endian::native)")
    with writer.scope():
        swap_fields(task, writer)


def encode_endian(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Write a struct method for encoding buffers."""

    with writer.javadoc():
        writer.write("Encode this instance to a buffer.")
        writer.empty()
        writer.write(
            task.command(
                "tparam",
                "    endianness Byte order for encoding elements.",
            )
        )
        writer.write(task.command("param[out]", "buffer     Buffer to write."))
        writer.write(
            task.command(
                "return", "               The number of bytes encoded."
            )
        )

    writer.write(f"template <std::endian endianness = {endian_str(task)}>")
    writer.write("inline std::size_t encode(Buffer *buffer) const")
    with writer.scope():
        writer.write("*buffer = *raw_ro();")

        with writer.padding():
            writer.write(
                f"reinterpret_cast<{task.name} *>"
                "(buffer)->endian<endianness>();"
            )

        writer.write("return size;")


def decode_endian(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Write a struct method for decoding buffers."""

    with writer.javadoc():
        writer.write("Update this instance from a buffer.")
        writer.empty()
        writer.write(
            task.command(
                "tparam",
                "   endianness Byte order from decoding elements.",
            )
        )
        writer.write(task.command("param[in]", "buffer     Buffer to read."))
        writer.write(
            task.command(
                "return", "              The number of bytes decoded."
            )
        )

    writer.write(f"template <std::endian endianness = {endian_str(task)}>")
    writer.write("inline std::size_t decode(const Buffer *buffer)")
    with writer.scope():
        writer.write("*raw() = *buffer;")

        with writer.padding():
            writer.write("endian<endianness>();")

        writer.write("return size;")


def struct_methods(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Write generated-struct methods."""

    writer.write("using Buffer = byte_array<size>;")
    writer.write("using Span = byte_span<size>;")
    with writer.padding():
        writer.write(f"auto operator<=>(const {task.name} &) const = default;")

    struct_buffer_method(task, writer, False)

    with writer.padding():
        span_method(task, writer)

    struct_buffer_method(task, writer, True)

    if task.instance["codec"]:
        writer.empty()
        handle_endian(task, writer)
        writer.empty()
        encode_endian(task, writer)
        writer.empty()
        decode_endian(task, writer)

    if task.instance["json"]:
        to_json_method(
            task,
            writer,
            protocol_json(task),
            dumps_indent=task.instance["json_indent"],
            task_name=False,
            static=True,
            definition=True,
        )



"""
A module for implementing struct stream-related methods.
"""

# third-party
from vcorelib.io.file_writer import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask


def struct_istream(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Generate an input-stream handling method."""

    writer.write(
        "inline byte_istream &operator>>"
        f"(byte_istream &stream, {task.name} &instance)"
    )

    with writer.scope():
        writer.write(
            f"stream.read(instance.raw()->data(), {task.name}::size);"
        )
        writer.write("return stream;")


def struct_ostream(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Generate an output-stream handling method."""

    writer.write(
        "inline byte_ostream &operator<<"
        f"(byte_ostream &stream, const {task.name} &instance)"
    )

    with writer.scope():
        writer.write(
            f"stream.write(instance.raw_ro()->data(), {task.name}::size);"
        )
        writer.write("return stream;")


def struct_stream_methods(
    task: GenerateTask, writer: IndentedFileWriter
) -> None:
    """Generate struct stream read and write methods."""

    writer.c_comment("Stream interfaces.")
    struct_istream(task, writer)
    struct_ostream(task, writer)



"""
A module implementing interfaces for working with ARM CMSIS-SVD files.
"""

# internal
from ifgen.svd.cpu import process_cpu
from ifgen.svd.device import process_device
from ifgen.svd.peripherals import process_peripheral, process_peripherals
from ifgen.svd.task import TAG_PROCESSORS


def register_processors() -> None:
    """Register tag-processing methods."""

    TAG_PROCESSORS["device"] = process_device
    TAG_PROCESSORS["cpu"] = process_cpu
    TAG_PROCESSORS["peripheral"] = process_peripheral
    TAG_PROCESSORS["peripherals"] = process_peripherals



"""
A module implementing a struct header-file interface generator.
"""

# built-in
from contextlib import ExitStack
from typing import Any

# third-party
from vcorelib.io.file_writer import (
    CommentStyle,
    IndentedFileWriter,
    LineWithComment,
)

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods import struct_methods
from ifgen.struct.methods.fields import bit_fields
from ifgen.struct.python import python_struct_header
from ifgen.struct.stream import struct_stream_methods

FieldConfig = dict[str, int | str]


def enforce_expected_size(
    size: int, data: dict[str, Any], assert_msg: str
) -> None:
    """Enforce an expected-size field."""

    # If expected size is set, verify it.
    if "expected_size" in data:
        assert data["expected_size"] == size, (
            assert_msg,
            data["expected_size"],
            size,
        )


def struct_line(
    name: str,
    value: FieldConfig,
    volatile: bool,
    const: bool,
    array_length: int = None,
    default: str = None,
) -> LineWithComment:
    """Build a string for a struct-field line."""

    line = f"{value['type']} {name}"
    if array_length is not None:
        line += f"[{name}_length]"

    prefix = "volatile " if volatile else ""
    prefix += "const " if const else ""

    if default:
        line += f" = {default}"
    elif const:
        line += " = {}"

    return prefix + f"{line};", value.get("description")  # type: ignore


def struct_fields(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Generate struct fields."""

    writer.c_comment("Fields.")

    with writer.trailing_comment_lines(style=CommentStyle.C_DOXYGEN) as lines:
        # Fields.
        for field in task.instance["fields"]:
            enforce_expected_size(
                task.env.size(field["type"]) * field.get("array_length", 1),
                field,
                f"{task.name}.{field['name']}",
            )

            if "array_length" in field:
                lines.append(
                    (
                        (
                            f"static constexpr std::size_t "
                            f"{field['name']}_length = "
                            f"{field['array_length']};"
                        ),
                        None,
                    )
                )

            with ExitStack() as stack:
                is_union = field.get("alternates")
                all_fields = [field]
                if is_union:
                    writer.write("union")
                    stack.enter_context(writer.scope(suffix=";"))
                    all_fields.extend(field["alternates"])

                for possible_union in all_fields:
                    if "type" not in possible_union:
                        possible_union["type"] = field["type"]

                    # Handle enum defaults.
                    default = None
                    if task.env.is_enum(possible_union["type"]):
                        enum = task.env.get_enum(possible_union["type"])
                        if enum.default:
                            default = f"{possible_union['type']}_default"
                            if "array_length" in field:
                                default = (
                                    "{"
                                    + ", ".join(
                                        [
                                            default
                                            for _ in range(
                                                field["array_length"]
                                            )
                                        ]
                                    )
                                    + "}"
                                )

                    line, comment = struct_line(
                        possible_union["name"],
                        possible_union,
                        possible_union["volatile"],
                        possible_union["const"],
                        array_length=possible_union.get("array_length"),
                        default=default,
                    )
                    if is_union:
                        writer.write(
                            line + ("" if not comment else f" /* {comment} */")
                        )
                    else:
                        lines.append((line, comment))

        lines.append(("", None))


def struct_instance(
    task: GenerateTask, writer: IndentedFileWriter, instance: dict[str, Any]
) -> None:
    """Generate struct instances."""

    writer.empty()

    if instance.get("description"):
        with writer.javadoc():
            writer.write(instance["description"])

    type_base = f"{task.name} *"
    writer.write(
        (
            "static "
            + ("volatile " if instance["volatile"] else "")
            + f"{type_base}const "
            f"{instance['name']} = "
            f"reinterpret_cast<{type_base}>({instance['address']});"
        )
    )


def cpp_struct_header(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Create the contents of a C++ struct header file."""

    attributes = ["gnu::packed"]
    writer.write(f"struct [[{', '.join(attributes)}]] {task.name}")
    with writer.scope(suffix=";"):
        writer.c_comment("Constant attributes.")
        with writer.trailing_comment_lines(
            style=CommentStyle.C_DOXYGEN
        ) as lines:
            if task.instance["identifier"]:
                lines.append(
                    (
                        "static constexpr struct_id_t "
                        f"id = {task.protocol().id};",
                        f"{task.name}'s identifier.",
                    )
                )

            # trace=True
            size = task.env.size(task.name)
            enforce_expected_size(size, task.instance, task.name)

            lines.append(
                (
                    f"static constexpr std::size_t size = {size};",
                    f"{task.name}'s size in bytes.",
                )
            )

        writer.empty()
        struct_fields(task, writer)

        # Methods.
        writer.c_comment("Methods.")

        if task.instance["methods"]:
            struct_methods(task, writer)

        bit_fields(task, writer)

    # Add size assertion.
    writer.empty()
    writer.write(f"static_assert(sizeof({task.name}) == {task.name}::size);")
    writer.write(f"static_assert(ifgen_struct<{task.name}>);")

    if task.instance["stream"]:
        writer.empty()
        struct_stream_methods(task, writer)

    for instance in task.instance.get("instances", []):
        struct_instance(task, writer, instance)


def struct_header(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Create the contents of a struct header file."""

    if task.is_cpp:
        cpp_struct_header(task, writer)
    elif task.is_python:
        python_struct_header(task, writer)



"""
Common utilities for generating bit-field related struct methods.
"""

# built-in
from typing import Any, Optional

BitField = dict[str, Any]


def bit_mask_literal(width: int) -> str:
    """Get a bit-mask literal."""
    return "0b" + ("1" * width) + "u"


def possible_array_arg(parent: dict[str, Any]) -> str:
    """Determine if a method needs an array-index argument."""

    array_length: Optional[int] = parent.get("array_length")
    inner = ""
    if array_length:
        inner = "std::size_t index"

    return inner


STANDARD_INTS = [
    ("uint8_t", 8),
    ("uint16_t", 16),
    ("uint32_t", 32),
    ("uint64_t", 64),
]


def bit_field_underlying(field: dict[str, Any]) -> str:
    """Get the underlying type for a bit field."""

    kind = field.get("type")

    # Automatically determine a sane primitive-integer type to use if one isn't
    # specified.
    if kind is None:
        width = field["width"]

        if width == 1:
            kind = "bool"
        else:
            for candidate, bit_width in STANDARD_INTS:
                if field["width"] <= bit_width:
                    kind = candidate
                    break

    assert kind is not None, kind
    return kind


def bit_field_method_slug(
    field: dict[str, Any], member: str = "", alias: str = None
) -> str:
    """Get a method slug for a struct's bit-field method."""

    name = str(field["name"]) if not alias else alias
    if member:
        name += "_" + member
    return name



"""
A module implementing an interface for generating bit-field methods for
structs.
"""

# built-in
from typing import Any

# third-party
from vcorelib.io.file_writer import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods.fields.common import BitField
from ifgen.struct.methods.fields.getter import (
    bit_field_get_all_method,
    bit_field_get_method,
)
from ifgen.struct.methods.fields.setter import (
    bit_field_set_all_method,
    bit_field_set_method,
)


def bit_field(
    task: GenerateTask,
    parent: dict[str, Any],
    field: BitField,
    writer: IndentedFileWriter,
    read_fields: list[BitField],
    write_fields: list[BitField],
    alias: str = None,
) -> None:
    """Generate for an individual bit-field."""

    type_size = task.env.size(parent["type"]) * 8

    index = field["index"]
    width = field["width"]

    # Validate field parameters.
    assert index + width <= type_size, (index, width, type_size, field)
    assert field["read"] or field["write"], field

    # Generate a 'get' method.
    if field["read"]:
        bit_field_get_method(task, parent, field, writer, alias=alias)
        read_fields.append(field)

    # Generate a 'set' method.
    if field["write"]:
        bit_field_set_method(task, parent, field, writer, alias=alias)
        write_fields.append(field)


def handle_atomic_fields_methods(
    task: GenerateTask,
    writer: IndentedFileWriter,
    field: dict[str, Any],
    read_fields: list[BitField],
    write_fields: list[BitField],
    alias: str = None,
) -> None:
    """Handle additional bit-field methods."""

    if len(read_fields) > 1:
        writer.empty()
        bit_field_get_all_method(task, writer, field, read_fields, alias=alias)

    if len(write_fields) > 1:
        writer.empty()
        bit_field_set_all_method(
            task, writer, field, write_fields, alias=alias
        )


def bit_fields(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Generate bit-field lines."""

    for field in task.instance["fields"]:
        read_fields: list[BitField] = []
        write_fields: list[BitField] = []

        for bfield in field.get("fields", []):
            bit_field(task, field, bfield, writer, read_fields, write_fields)
        handle_atomic_fields_methods(
            task, writer, field, read_fields, write_fields
        )

        for alternate in field.get("alternates", []):
            read_fields = []
            write_fields = []

            for bfield in alternate.get("fields", []):
                bit_field(
                    task,
                    field,
                    bfield,
                    writer,
                    read_fields,
                    write_fields,
                    alias=alternate["name"],
                )
            handle_atomic_fields_methods(
                task,
                writer,
                field,
                read_fields,
                write_fields,
                alias=alternate["name"],
            )



"""
A module implementing interfaces to facilitate code generation.
"""

# built-in
from importlib import import_module
from logging import getLogger
from multiprocessing.pool import ThreadPool
from pathlib import Path
import sys

# third-party
from vcorelib.names import import_str_and_item

# internal
from ifgen.common import create_common, create_common_test
from ifgen.config import Config
from ifgen.enum import create_enum, create_enum_source, create_enum_test
from ifgen.enums import Generator, Language
from ifgen.environment import IfgenEnvironment
from ifgen.generation.interface import GenerateTask, InstanceGenerator
from ifgen.plugins import internal_plugin_entry
from ifgen.struct import (
    create_struct,
    create_struct_test,
)

CodeGenerators = dict[Generator, list[InstanceGenerator]]
GENERATORS: CodeGenerators = {
    Generator.STRUCTS: [create_struct, create_struct_test],
    Generator.ENUMS: [create_enum, create_enum_test, create_enum_source],
    Generator.IFGEN: [],
    Generator.CUSTOM: [internal_plugin_entry],
}
LOG = getLogger(__name__)


def resolve_generators(env: IfgenEnvironment) -> CodeGenerators:
    """
    Populate any custom generator interfaces defined in the configuration.
    """

    path = str(env.root_path.resolve())
    if path not in sys.path:
        sys.path.append(path)

    generators = GENERATORS.copy()

    if env.config.data["common"]:
        generators[Generator.IFGEN].append(create_common)
    if env.config.data["common_test"]:
        generators[Generator.IFGEN].append(create_common_test)

    for custom in env.config.data.get("plugins", []):
        module, app = import_str_and_item(custom)
        generators[Generator.CUSTOM].append(
            getattr(import_module(module), app)
        )

    return generators


def generate(root: Path, config: Config) -> None:
    """Generate struct files."""

    env = IfgenEnvironment(root, config)

    # Search for language configurations.
    languages = [x for x in Language if config.data.get(x.cfg_dir_name)]
    LOG.info("Generating for languages: %s.", languages)

    with ThreadPool() as pool:
        for language in languages:
            for generator, methods in resolve_generators(env).items():
                for method in methods:
                    pool.map(
                        method,
                        (
                            GenerateTask(
                                name,
                                generator,
                                language,
                                env.make_path(
                                    name,
                                    generator,
                                    language,
                                    from_output=True,
                                ),
                                env.make_test_path(name, generator, language),
                                data,
                                env,
                            )
                            for name, data in config.data.get(
                                generator.value, {}
                            ).items()
                        ),
                    )

    env.prune_empty()



"""
Test that we can load test resources.
"""

# built-in
from pathlib import Path
from tempfile import TemporaryDirectory

# module under test
from ifgen.generation.interface import apply_cpp_namespace
from ifgen.paths import audit_init_file

# internal
from tests.resources import resource


def test_resources_basic():
    """Test that we can locate test data."""

    assert apply_cpp_namespace("test", "test", header=False)

    assert resource("test.txt").is_file()
    assert resource("test.txt", valid=False).is_file()

    with TemporaryDirectory() as tmpdir:
        audit_init_file(Path(tmpdir, "test.py"))



"""
A module implementing code-generation interfaces for Python.
"""

# built-in
from contextlib import contextmanager
from typing import Iterable, Iterator

# third-party
from vcorelib.io import IndentedFileWriter

PythonImports = dict[str, list[str]]


def write_imports(
    writer: IndentedFileWriter,
    label: str,
    imports: PythonImports,
) -> None:
    """Write a group of Python imports to a file."""

    writer.empty()
    writer.write(f"# {label}")
    for key, values in sorted(imports.items()):
        writer.write(f"from {key} import {', '.join(values)}")


def python_imports(
    writer: IndentedFileWriter,
    built_in: PythonImports = None,
    third_party: PythonImports = None,
    internal: PythonImports = None,
    final_empty: int = 2,
) -> None:
    """Write Python-module imports."""

    if built_in:
        write_imports(writer, "built-in", built_in)
    if third_party:
        write_imports(writer, "third-party", third_party)
    if internal:
        write_imports(writer, "internal", internal)

    writer.empty(count=final_empty)


def python_docstring(
    writer: IndentedFileWriter, *docstrings: str, empty: bool = True
) -> None:
    """Write docstring data."""

    writer.write('"""')
    for docstring in docstrings:
        writer.write(docstring)
    writer.write('"""')
    if empty:
        writer.empty()


@contextmanager
def python_class(
    writer: IndentedFileWriter,
    name: str,
    docstring: str,
    parents: list[str] = None,
    final_empty: int = 2,
) -> Iterator[None]:
    """Write class definition contents."""

    line = f"class {name}"
    if parents:
        line += f"({', '.join(parents)})"
    line += ":"

    writer.write(line)
    with writer.indented():
        python_docstring(writer, docstring)
        yield

    writer.empty(count=final_empty)


@contextmanager
def python_function(
    writer: IndentedFileWriter,
    name: str,
    docstring: str,
    params: str = "",
    return_type: str = "None",
    final_empty: int = 2,
    decorators: Iterable[str] = None,
) -> Iterator[None]:
    """Write a Python function."""

    if decorators:
        for decorator in decorators:
            writer.write("@" + decorator)
    writer.write(f"def {name}({params}) -> {return_type}:")

    with writer.indented():
        python_docstring(writer, docstring)
        yield

    writer.empty(count=final_empty)



"""
A module implementing a generation-environment interface.
"""

# built-in
from pathlib import Path
from typing import Any, NamedTuple, Optional

# third-party
from runtimepy.codec.protocol import Protocol
from runtimepy.codec.system import TypeSystem
from runtimepy.enum import RuntimeEnum
from vcorelib.logging import LoggerMixin
from vcorelib.names import to_snake
from vcorelib.paths import normalize, prune_empty_directories
from vcorelib.paths.context import TextPreprocessor

# internal
from ifgen import PKG_NAME
from ifgen.config import Config
from ifgen.enums import Generator, Language
from ifgen.environment.field import process_field
from ifgen.environment.padding import PaddingManager, type_string
from ifgen.paths import combine_if_not_absolute, create_formatter


def runtime_enum_data(data: dict[str, Any]) -> dict[str, int]:
    """Get runtime enumeration data."""

    result = {}

    curr_value = 0

    for key, value in data.items():
        if value is None or "value" not in value:
            result[key] = curr_value
            curr_value += 1
        else:
            result[key] = value["value"]
            if value["value"] >= curr_value:
                curr_value = value["value"] + 1

    return result


class Directories(NamedTuple):
    """A collection of directories relevant to code generation outputs."""

    config_parts: list[str]
    source: Path
    output: Path
    test_dir: Path

    def prune_empty(self) -> None:
        """Attempt to eliminate any empty output directories."""

        for path in (self.source, self.output, self.test_dir):
            prune_empty_directories(path)


class IfgenEnvironment(LoggerMixin):
    """A class for managing stateful information while generating outputs."""

    def __init__(self, root: Path, config: Config) -> None:
        """Initialize this instance."""

        super().__init__()
        self.root_path = root
        self.config = config

        # Load per-language directories.
        self.directories: dict[Language, Directories] = {}
        for language in Language:
            result = self.get_dirs(language)
            if result is not None:
                self.directories[language] = result

        self.generated: set[Path] = set()

        # Create output directories.
        for language, dirs in self.directories.items():
            for subdir in Generator:
                for path in [dirs.output, dirs.test_dir]:
                    path.joinpath(subdir).mkdir(parents=True, exist_ok=True)

        self.types = TypeSystem(
            *(
                self.config.data["namespace"]
                if self.config.data["namespace"]
                else [PKG_NAME.capitalize()]
            )
        )
        self.padding = PaddingManager()
        self._register_enums()
        self._register_structs()

        # Set up language-specific preprocessors.
        self.preprocessors: dict[Language, TextPreprocessor] = {
            Language.CPP: create_formatter(
                self.config.data["clang_format"], cwd=self.root_path
            )
        }

    def prune_empty(self) -> None:
        """Attempt to eliminate any empty output directories."""

        for dirs in self.directories.values():
            dirs.prune_empty()

    def get_dirs(self, langauge: Language) -> Optional[Directories]:
        """Get source, output and test directories."""

        result = None

        cfg_dir = langauge.cfg_dir_name
        if cfg_dir in self.config.data:
            dirs = self.config.data[cfg_dir]
            if dirs:
                source = combine_if_not_absolute(
                    self.root_path, normalize(*dirs)
                )
                output = combine_if_not_absolute(
                    source,
                    (
                        normalize(*self.config.data["output_dir"])
                        if self.config.data["output_dir"]
                        else ""
                    ),
                )
                test_dir = combine_if_not_absolute(
                    source,
                    (
                        normalize(*self.config.data["test_dir"])
                        if self.config.data["test_dir"]
                        else ""
                    ),
                )
                result = Directories(dirs, source, output, test_dir)

        return result

    def _register_enums(self) -> None:
        """Register configuration enums."""

        for name, enum in self.config.data.get("enums", {}).items():
            self.types.enum(
                name,
                runtime_enum_data(enum["enum"]),
                *enum["namespace"],
                primitive=type_string(enum["underlying"]),
                default=enum.get("default"),
            )

            self.logger.info(
                "Registered enum '%s'.",
                self.types.root_namespace.delim.join(
                    enum["namespace"] + [name]
                ),
            )

    def _register_structs(self) -> None:
        """Register configuration structs."""

        for name, struct in self.config.data.get("structs", {}).items():
            namespace = [*struct["namespace"]]
            self.types.register(name, *namespace)

            field_groups = []

            self.padding.reset()
            for field in struct["fields"]:
                padding = list(
                    process_field(
                        name, self.padding, self.types, field, namespace
                    )
                )
                if padding:
                    field_groups.append(padding)
                field_groups.append([field])

            # Re-assign fields structure.
            struct["fields"] = []
            for group in field_groups:
                for field in group:
                    struct["fields"].append(field)

            self.logger.info(
                "Registered struct '%s' (%d bytes).",
                self.types.root_namespace.delim.join(
                    struct["namespace"] + [name]
                ),
                self.types.size(name, *struct["namespace"]),
            )

    def make_path(
        self,
        name: str,
        generator: Generator,
        language: Language,
        from_output: bool = False,
        track: bool = True,
    ) -> Path:
        """Make part of a task's path."""

        if language is Language.PYTHON:
            name = to_snake(name)

        result = Path(str(generator), f"{name}.{language.header_suffix}")

        if from_output:
            result = self.directories[language].output.joinpath(result)

        if track:
            self.generated.add(result)

        return result

    def make_test_path(
        self, name: str, generator: Generator, language: Language
    ) -> Path:
        """Make a path to an interface's unit-test suite."""

        result = self.directories[language].test_dir.joinpath(
            str(generator), f"test_{name}.{language.source_suffix}"
        )
        self.generated.add(result)
        return result

    def rel_include(
        self, name: str, generator: Generator, language: Language
    ) -> Path:
        """Get an include path to a generated output."""

        return (
            self.directories[language]
            .output.relative_to(self.directories[language].source)
            .joinpath(self.make_path(name, generator, language, track=False))
        )

    def get_protocol(self, name: str, exact: bool = False) -> Protocol:
        """Get the protocol instance for a given struct."""

        return self.types.get_protocol(
            name, self.config.data["structs"].get("namespace", []), exact=exact
        )

    def is_struct(self, name: str) -> bool:
        """Determine if a field is a struct or not."""

        try:
            self.get_protocol(type_string(name))
            return True
        except KeyError:
            return False

    def size(
        self, type_name: str, exact: bool = False, trace: bool = False
    ) -> int:
        """Get the size of a given type."""
        return self.types.size(
            type_string(type_name), exact=exact, trace=trace
        )

    def get_enum(self, name: str, exact: bool = False) -> RuntimeEnum:
        """Get a runtime enum instance for a given enumeration."""

        return self.types.get_enum(
            name, *self.config.data["enums"].get("namespace", []), exact=exact
        )

    def is_enum(self, name: str, exact: bool = False) -> bool:
        """Determine if a field is an enumeration or not."""

        try:
            self.get_enum(type_string(name), exact=exact)
            return True
        except KeyError:
            return False



"""
A module implementing interfaces for string-data elements.
"""

# built-in
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Iterable, Optional, Type, TypeVar
from xml.etree import ElementTree

# third-party
from vcorelib.logging import LoggerType


@dataclass
class StringKeyVal:
    """Name of key and whether or not it's required."""

    key: str
    required: bool = True


T = TypeVar("T", bound="StringKeyValueMixin")


class StringKeyValueMixin(ABC):
    """A mixin for SVD data model classes."""

    raw_data: dict[str, str]

    @classmethod
    def create(cls: Type[T], elem: ElementTree.Element, *args, **kwargs) -> T:
        """Create a device instance from an element."""

        inst = cls(*args, **kwargs)
        inst.raw(elem)
        return inst

    @classmethod
    @abstractmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

    def handle_description(
        self, data: dict[str, Any] = None, prefix: str = None
    ) -> dict[str, Any]:
        """Handle a possible description entry."""

        if data is None:
            data = {}

        description = self.raw_data.get("description")

        if description:
            data["description"] = (prefix if prefix else "") + description

        return data

    def raw(self, elem: ElementTree.Element) -> dict[str, str]:
        """Get raw data for this instance based on string keys."""

        result = getattr(self, "raw_data", None)
        if result is None:
            self.raw_data = self.get_values(elem)
            result = self.raw_data

        assert result is not None
        return result

    def log(
        self, elem: ElementTree.Element, logger: Optional[LoggerType]
    ) -> None:
        """Log information from this instance's raw data."""

        if logger is not None:
            for key, value in self.raw(elem).items():
                for line in value.splitlines():
                    line = line.strip().replace("\\n", "")
                    if line:
                        logger.info("%s: %s", key, line)

    @classmethod
    def get_values(cls, elem: ElementTree.Element) -> dict[str, str]:
        """Get string values for this instance."""

        return get_string_values(elem, cls.string_keys())


def get_string_values(
    elem: ElementTree.Element, keys: Iterable[StringKeyVal]
) -> dict[str, str]:
    """Get string values from an element's children."""

    result = {}

    for item in keys:
        key = item.key
        required = item.required

        new_elem = elem.find(key)
        assert (
            new_elem is not None or not required
        ), f"'{key}' required but not found!"
        if new_elem is not None:
            result[key] = new_elem.text or ""

    return result



"""
A module for generating static map data structures related to enumerations.
"""

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask


def enum_to_string_map(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Create a static map for enum to string conversion."""

    inst = f"{task.name}_to_string"
    writer.write(
        f"static const std::map<{task.name}, const char *> {inst} = " + "{"
    )

    for enum in task.instance.get("enum", {}):
        with writer.indented():
            writer.write("{" + f'{task.name}::{enum}, "{enum}"' + "},")

    writer.write("};")


def enum_from_string_map(
    task: GenerateTask, writer: IndentedFileWriter
) -> None:
    """Create a static map for string to enum conversion."""

    inst = f"{task.name}_from_string"
    writer.write(
        f"static const std::map<const std::string, {task.name}> {inst} = "
        + "{"
    )

    for enum in task.instance.get("enum", {}):
        with writer.indented():
            writer.write("{" + f'"{enum}", {task.name}::{enum}' + "},")

    writer.write("};")



"""
A module implementing a data model for ARM CMSIS-SVD 'device' data.
"""

# built-in
from dataclasses import dataclass
from typing import Iterable

# internal
from ifgen.svd.string import StringKeyVal, StringKeyValueMixin

# https://www.keil.com/pack/doc/CMSIS/SVD/html/
#     elem_special.html#registerPropertiesGroup_gr
REGISTER_PROPERTIES = [
    StringKeyVal("size", False),
    StringKeyVal("access", False),
    StringKeyVal("protection", False),
    StringKeyVal("resetValue", False),
    StringKeyVal("resetMask", False),
]
ARRAY_PROPERTIES = [
    StringKeyVal("dim", False),
    StringKeyVal("dimIncrement", False),
    StringKeyVal("dimIndex", False),
    StringKeyVal("dimName", False),
    # This is an object.
    # StringKeyVal("dimArrayIndex", False),
]


@dataclass
class Device(StringKeyValueMixin):
    """A container for device information."""

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        # https://www.keil.com/pack/doc/CMSIS/SVD/html/elem_device.html
        return [
            StringKeyVal("vendor", False),
            StringKeyVal("vendorID", False),
            StringKeyVal("name", True),
            StringKeyVal("series", False),
            StringKeyVal("version", True),
            StringKeyVal("description", True),
            StringKeyVal("licenseText", False),
            StringKeyVal("headerSystemFilename", False),
            StringKeyVal("headerDefinitionsPrefix", False),
            StringKeyVal("addressUnitBits", True),
            StringKeyVal("width", True),
        ] + REGISTER_PROPERTIES



"""
An interface for optional code generation plugins.
"""

# internal
from ifgen.enums import Language
from ifgen.generation.interface import GenerateTask
from ifgen.plugins.struct_receiver import (
    cpp_struct_receiver,
    python_struct_receiver,
)

PLUGINS = {
    "struct_receiver": {
        Language.PYTHON: python_struct_receiver,
        Language.CPP: cpp_struct_receiver,
    }
}


def internal_plugin_entry(task: GenerateTask) -> None:
    """A generator for a struct receiver interface."""

    if task.name in PLUGINS and task.language in PLUGINS[task.name]:
        PLUGINS[task.name][task.language](task)



"""
A module defining generator interfaces.
"""

# built-in
from contextlib import ExitStack, contextmanager
from json import dumps
from pathlib import Path
from typing import Any, Callable, Iterable, Iterator, NamedTuple, Optional

# third-party
from runtimepy.codec.protocol import Protocol
from runtimepy.enum import RuntimeEnum
from vcorelib.io import IndentedFileWriter
from vcorelib.namespace import CPP_DELIM
from vcorelib.paths.context import TextPreprocessor

# internal
from ifgen import PKG_NAME, VERSION
from ifgen.enums import Generator, Language
from ifgen.environment import IfgenEnvironment
from ifgen.paths import audit_init_file

InstanceConfig = dict[str, Any]
IfgenConfig = dict[str, Any]


class TypeLookup(NamedTuple):
    """A container for type-lookup results."""

    name: str
    final: str
    generator: Generator


def apply_cpp_namespace(
    name: str, data: str, header: bool = True, prefix: str = None
) -> str:
    """Namespace a string with this task's name."""

    if not header:
        data = f"{name}::{data}"
    return data if not prefix else prefix + data


@contextmanager
def ifndef_guard(writer: IndentedFileWriter, text: str) -> Iterator[None]:
    """Guard contents with header macro convention."""

    writer.write(f"#ifndef {text}")
    writer.write(f"#define {text}")
    try:
        yield
    finally:
        writer.empty()
        writer.write("#endif")


class GenerateTask(NamedTuple):
    """A container for instance-generation tasks."""

    name: str
    generator: Generator
    language: Language

    path: Path
    test_path: Path

    instance: InstanceConfig
    env: IfgenEnvironment

    def cpp_namespace(
        self, data: str, header: bool = True, prefix: str = None
    ) -> str:
        """Namespace a string with this task's name."""

        return apply_cpp_namespace(
            self.name, data, header=header, prefix=prefix
        )

    @property
    def is_python(self) -> bool:
        """Determine if this task's language is set to Python."""
        return self.language is Language.PYTHON

    @property
    def is_cpp(self) -> bool:
        """Determine if this task's language is set to C++."""
        return self.language is Language.CPP

    @property
    def stream_implementation(self) -> bool:
        """
        Determine if this instances should include a stream implementations.
        """
        return self.instance.get("stream", True)  # type: ignore

    @property
    def source_path(self) -> Path:
        """Get a source file for this task."""
        return self.path.with_suffix(f".{self.language.source_suffix}")

    @property
    def config(self) -> IfgenConfig:
        """Get the environment's configuration data."""
        return self.env.config.data

    def enum(self) -> RuntimeEnum:
        """Look up a runtime enumeration for this task."""

        return self.env.get_enum(self.name)

    def protocol(self) -> Protocol:
        """Loop up a protocol for this task."""

        return self.env.get_protocol(self.name)

    def namespace(self, *names: str) -> str:
        """Get this task's namespace."""

        nspace = self.env.types.root_namespace
        with nspace.pushed(*self.instance.get("namespace", []), *names):
            result = nspace.namespace(track=False)

        assert result, f"No namespace for '{self.name}'!"
        return result

    def check_custom_type(self, name: str) -> Optional[TypeLookup]:
        """Check if a name refers to a custom type."""

        final = name.split(CPP_DELIM)[-1]

        result = None
        for gen in Generator:
            if final in self.config.get(gen.value, {}):
                result = TypeLookup(name, final, gen)
                break
        return result

    def custom_include(self, name: str) -> Optional[Path]:
        """Attempt to build a path to a custom include."""

        result = None

        lookup = self.check_custom_type(name)
        if lookup is not None:
            result = (
                Path(
                    "..",
                    self.env.make_path(
                        lookup.final, lookup.generator, self.language
                    ),
                )
                if lookup.generator != self.generator
                else Path(f"{lookup.final}.{self.language.header_suffix}")
            )

        return result

    def command(self, command: str, data: str = "", space: str = " ") -> str:
        """Get a doxygen command string."""
        return (
            str(self.config["command"])
            + command
            + (space if data else "")
            + data
        )

    @contextmanager
    def source_boilerplate(
        self, includes: Iterable[str]
    ) -> Iterator[IndentedFileWriter]:
        """Create standard generation boilerplate for a source file."""

        with ExitStack() as stack:
            writer = stack.enter_context(
                IndentedFileWriter.from_path_if_different(
                    self.source_path,
                    per_indent=4,
                    preprocessor=self._text_preprocessor(),
                )
            )

            self.javadoc_header(writer)

            include = self.env.rel_include(
                self.name, self.generator, self.language
            )
            self.write_includes(
                writer,
                includes=[f'"{include}"'] + list(includes),
            )

            self.handle_namespace(stack, writer)

            yield writer

    def javadoc_header(
        self, writer: IndentedFileWriter, json: bool = False
    ) -> None:
        """Write a standard file header."""

        description = f"Generated by {PKG_NAME} ({VERSION})."

        with ExitStack() as stack:
            if self.is_python:
                stack.enter_context(
                    writer.scope(opener='"""', closer='"""', indent=0)
                )
                writer.write(description)
            else:
                stack.enter_context(writer.javadoc())
                writer.write(self.command("file"))
                writer.write(self.command("brief", description))

            if json:
                writer.write(dumps(self.instance, indent=2))

    def handle_namespace(
        self, stack: ExitStack, writer: IndentedFileWriter
    ) -> None:
        """Write namespace boilerplate to a file."""

        if self.is_python:
            return

        # Write namespace.
        namespace = self.namespace()
        writer.write(f"namespace {namespace}")

        stack.enter_context(
            writer.scope(suffix=f"; // namespace {namespace}", indent=0)
        )

        stack.enter_context(writer.padding())

    def write_includes(
        self, writer: IndentedFileWriter, includes: Iterable[str] = None
    ) -> None:
        """Write sorted includes to a file."""

        if self.is_python:
            return

        with writer.padding():
            for include in sorted(includes if includes else []):
                if include:
                    writer.write(f"#include {include}")

    def resolve_description(
        self, description: Optional[str] = None
    ) -> Optional[str]:
        """Attempt to resolve a config-driven description."""

        if description is None:
            if "description" in self.instance and self.instance["description"]:
                description = self.instance["description"]

        return description

    def _text_preprocessor(self) -> Optional[TextPreprocessor]:
        """
        Get a text preprocessing method (i.e. code formatting) for this task.
        """
        return self.env.preprocessors.get(self.language)

    @contextmanager
    def boilerplate(
        self,
        includes: Iterable[str] = None,
        is_test: bool = False,
        use_namespace: bool = True,
        description: Optional[str] = None,
        json: bool = False,
        parent_depth: int = 0,
    ) -> Iterator[IndentedFileWriter]:
        """
        Create standard generation boilerplate and yield the file writer to
        use for writing the remaining content.
        """

        path = self.path if not is_test else self.test_path

        if self.is_python:
            audit_init_file(path, parent_depth=parent_depth)

        with ExitStack() as stack:
            writer = stack.enter_context(
                IndentedFileWriter.from_path_if_different(
                    path, per_indent=4, preprocessor=self._text_preprocessor()
                )
            )

            # Write file header.
            self.javadoc_header(writer, json=json)

            if self.is_cpp:
                writer.empty()
                if not is_test:
                    writer.write("#pragma once")
                    stack.enter_context(
                        ifndef_guard(
                            writer,
                            "_".join(
                                [
                                    x.upper()
                                    for x in [
                                        self.namespace().replace("::", "_"),
                                        path.parent.name,
                                        self.path.name.replace(".", "_"),
                                    ]
                                ]
                            ),
                        )
                    )
                else:
                    writer.write("#ifdef NDEBUG")
                    writer.write("#undef NDEBUG")
                    writer.write("#endif")

            self.write_includes(writer, includes=includes)

            if use_namespace:
                self.handle_namespace(stack, writer)

            if not self.is_python:
                description = self.resolve_description(description=description)
                if description:
                    with writer.javadoc():
                        writer.write(description)

            yield writer

    def has_instances(self) -> bool:
        """Determine whether or not this instance has concrete instances."""
        return bool(self.instance.get("instances", []))

    def method_suffix(self) -> str:
        """Get a possible suffix for a method."""
        return "" if not self.has_instances() else " volatile"


InstanceGenerator = Callable[[GenerateTask], None]



"""
A module implementing interfaces for processing a group of peripherals.
"""

# built-in
from pathlib import Path
from typing import Any

# third-party
from vcorelib.io import ARBITER

# internal
from ifgen.svd.group.base import PeripheralGroup, peripheral_groups
from ifgen.svd.group.fields import EnumMap, StructMap, struct_fields
from ifgen.svd.model.peripheral import Peripheral

__all__ = ["PeripheralGroup", "peripheral_groups", "handle_group"]


def struct_instance(peripheral: Peripheral) -> dict[str, Any]:
    """Get struct instance data."""

    return {
        "name": peripheral.name,
        "address": peripheral.raw_data["baseAddress"],
    }


def struct_data(
    group: PeripheralGroup,
    structs: StructMap,
    enums: EnumMap,
    min_enum_members: int,
) -> dict[str, Any]:
    """Get struct data for a peripheral group."""

    data: dict[str, Any] = {}
    peripheral = group.root
    peripheral.handle_description(data)

    data["instances"] = [struct_instance(x) for x in group.peripherals]
    size, data["fields"] = struct_fields(
        peripheral.registers,
        structs,
        enums,
        peripheral.base_name(lower=False),
        min_enum_members,
    )

    # Too difficult due to padding.
    # data["expected_size"] = size
    del size

    return data


def handle_group(
    output_dir: Path,
    group: PeripheralGroup,
    includes: set[Path],
    min_enum_members: int,
) -> None:
    """Handle a peripheral group."""

    output = output_dir.joinpath("include.yaml")
    includes.add(output)

    structs: StructMap = {}
    enums: EnumMap = {}
    structs[group.root.base_name()] = struct_data(
        group, structs, enums, min_enum_members
    )
    ARBITER.encode(output, {"structs": structs, "enums": enums})



"""
A module for handling SVD bit-field enumerations.
"""

# built-in
from os.path import commonprefix
from typing import Any

# internal
from ifgen.svd.model.enum import EnumeratedValues

EnumValues = dict[str, Any]
ENUM_DEFAULTS: dict[str, Any] = {
    "unit_test": False,
    "json": False,
    "use_map": False,
    "identifier": False,
}

BY_HASH: dict[str, dict[int, str]] = {}
PRUNE_ENUMS = False


def get_enum_name(name: str, peripheral: str, raw_mapping: EnumValues) -> str:
    """Get the name of an enumeration."""

    if not PRUNE_ENUMS:
        return name

    hashed = hash(
        ",".join(
            name + f"={val['value']}"
            for name, val in sorted(raw_mapping.items())
        )
    )

    BY_HASH.setdefault(peripheral, {})

    for_periph = BY_HASH[peripheral]
    for_periph.setdefault(hashed, name)

    return for_periph[hashed]


IGNORE_WORDS = {
    "the",
    "as",
    "a",
    "is",
    "will",
    "but",
    "are",
    "yet",
    "that",
    "to",
    "and",
    "in",
    "of",
    "on",
    "for",
    "from",
    "its",
    "it",
}


def is_name_part(value: str) -> bool:
    """Determine if a word should be part of an enumeration value name."""
    return bool(value) and value not in IGNORE_WORDS


def as_alnum(word: str) -> str:
    """Get a word's alpha-numeric contents only."""

    result = ""
    for char in word:
        if char.isalnum() or char == "_":
            result += char

    return result


SKIP = {"-"}


def handle_enum_name(name: str, description: str = None) -> str:
    """Attempt to generate more useful enumeration names."""

    if name.startswith("value") and description:
        alnum_parts = [
            as_alnum(x.strip().lower().replace("-", "_"))
            for x in description.split()
            if x not in SKIP
        ]

        # Prune some words if the description is very long.
        if len(alnum_parts) > 1:
            alnum_parts = list(filter(is_name_part, alnum_parts))

        assert alnum_parts, (name, description)

        new_name = "_".join(alnum_parts)

        assert new_name, (name, description)
        name = new_name

    return name


def remove_common_prefixes(data: dict[str, Any]) -> dict[str, Any]:
    """Attempt to remove common prefixes in enumeration names."""

    result = data

    length = len(commonprefix(list(result)))
    if length > 1:
        result = {
            key[length:] if length < len(key) else key: value
            for key, value in result.items()
        }

    return result


def handle_duplicate(existing: dict[str, Any], key: str, value: Any) -> None:
    """Handle key de-duplication for enumerations."""

    assert key
    if key[0].isnumeric():
        key = "_" + key

    while key in existing:
        key += "_x"

    existing[key] = value


def translate_enums(enum: EnumeratedValues) -> EnumValues:
    """Generate an enumeration definition."""

    result: dict[str, Any] = {}
    enum.handle_description(result)

    for name, value in enum.derived_elem.enum.items():
        enum_data: dict[str, Any] = {}
        value.handle_description(enum_data)

        value_str: str = value.raw_data["value"].lower()

        prefix = ""
        for possible_prefix in ("#", "0b", "0x"):
            if value_str.startswith(possible_prefix):
                prefix = possible_prefix
                break

        if prefix in ("#", "0b"):
            enum_data["value"] = int(
                value_str[len(prefix) :].replace("x", "1"), 2
            )
        elif prefix == "0x":
            enum_data["value"] = int(value_str[len(prefix) :], 16)
        else:
            enum_data["value"] = int(value_str)

        handle_duplicate(
            result,
            handle_enum_name(name, value.raw_data.get("description")),
            enum_data,
        )

    # Truncate names.
    new_result: dict[str, Any] = {}
    for key, value in remove_common_prefixes(result).items():
        handle_duplicate(
            new_result, key if len(key) < 51 else key[:45] + "_cont", value
        )

    return new_result



"""
A module implementing Python enum-generation interfaces.
"""

# third-party
from runtimepy.enum.registry import DEFAULT_ENUM_PRIMITIVE
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.generation.python import (
    python_class,
    python_docstring,
    python_function,
    python_imports,
)


def strip_t_suffix(data: str) -> str:
    """Strip a possible '_t' suffix from a string."""
    return data.replace("_t", "") if data.endswith("_t") else data


def uses_auto(task: GenerateTask) -> bool:
    """
    Determine if a generation task will require 'auto' from the enum module.
    """

    result = False

    for value in task.instance.get("enum", {}).values():
        if not value or "value" not in value:
            result = True
            break

    return result


def to_enum_name(data: str) -> str:
    """Ensure a candidate name string is suitable as an enumeration name."""
    return data.upper()


def python_enum_header(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Create a Python module for an enumeration."""

    built_in = {}
    if uses_auto(task):
        built_in["enum"] = ["auto"]
    built_in["typing"] = ["Optional"]

    runtime = task.enum()

    # Write imports.
    imports = ["RuntimeIntEnum"]
    if runtime.default:
        imports.append("T")
    python_imports(
        writer,
        third_party={"runtimepy.enum.registry": imports},
        built_in=built_in,
    )

    with python_class(
        writer,
        task.name,
        task.resolve_description() or "No description.",
        parents=["RuntimeIntEnum"],
        final_empty=0,
    ):
        # Write values.
        for enum, value in task.instance.get("enum", {}).items():
            final = "auto()"
            if value:
                final = value.get("value", final)

            writer.write(f"{to_enum_name(enum)} = {final}")
            if value and "description" in value:
                python_docstring(writer, value["description"])

        # Override underlying primitive if necessary.
        underlying = strip_t_suffix(task.instance["underlying"])
        if underlying != DEFAULT_ENUM_PRIMITIVE:
            with python_function(
                writer,
                "primitive",
                "The underlying primitive type for this runtime enumeration.",
                params="cls",
                return_type="str",
                final_empty=1,
                decorators=["classmethod"],
            ):
                writer.write(f'return "{underlying}"')

        # Write identifier.
        with python_function(
            writer,
            "id",
            "Get this enumeration's integer identifier.",
            params="cls",
            return_type="Optional[int]",
            final_empty=0,
            decorators=["classmethod"],
        ):
            writer.write(f"return {runtime.id}")

        if runtime.default:
            writer.empty()
            with python_function(
                writer,
                "default",
                "Get a possible default value for this enumeration.",
                params="cls: type[T]",
                return_type="Optional[T]",
                final_empty=0,
                decorators=["classmethod"],
            ):
                writer.write(
                    f"return cls.normalize(\"{task.instance['default']}\")"
                )



"""
A module implementing a 'peripherals' SVD-element processor.
"""

# built-in
from xml.etree import ElementTree

# third-party
from vcorelib.logging import LoggerType

# internal
from ifgen.svd.model.derived import derived_from_stack
from ifgen.svd.model.peripheral import Peripheral
from ifgen.svd.process import handle_registers
from ifgen.svd.task import SvdProcessingTask


def process_peripheral(
    elem: ElementTree.Element, task: SvdProcessingTask, logger: LoggerType
) -> None:
    """Process a SVD peripheral element."""

    derived_periph = None
    derived = elem.attrib.get("derivedFrom")
    if derived is not None:
        derived_periph = task.model.peripherals[derived]

    peripheral = Peripheral.create(elem, derived_periph, [], [], [])
    peripheral.log(elem, logger)

    # Handle registers.
    registers = elem.find("registers")
    if registers is not None:
        peripheral.registers = handle_registers(registers, peripheral)

    # Handle address blocks.
    for address_block in elem.iterfind("addressBlock"):
        peripheral.handle_address_block(address_block)

    # Handle interrupts.
    for interrupt in elem.iterfind("interrupt"):
        peripheral.handle_interrupt(interrupt)

    task.model.register_peripheral(elem, peripheral)


def process_peripherals(
    elem: ElementTree.Element, task: SvdProcessingTask, logger: LoggerType
) -> None:
    """Process a SVD peripherals element."""

    del logger

    for item in derived_from_stack(elem.iterfind("peripheral")):
        task.process(item)



"""
A module implementing a configuration interface for the package.
"""

# built-in
from typing import Any

# third-party
from vcorelib.dict import merge
from vcorelib.dict.codec import BasicDictCodec as _BasicDictCodec
from vcorelib.io import ARBITER as _ARBITER
from vcorelib.io import DEFAULT_INCLUDES_KEY
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.paths import Pathlike, find_file

# internal
from ifgen import PKG_NAME
from ifgen.schemas import IfgenDictCodec


class Config(IfgenDictCodec, _BasicDictCodec):
    """The top-level configuration object for the package."""

    def init(self, data: _JsonObject) -> None:
        """Initialize this instance."""

        super().init(data)

        common = ["identifier", "unit_test"]

        # Forward enum settings.
        enum_forwards = common + ["use_map"]
        enum: dict[str, Any]
        for enum in data.get("enums", {}).values():  # type: ignore
            for forward in enum_forwards:
                enum.setdefault(
                    forward,
                    data["enum"][forward],  # type: ignore
                )

        # Forward struct settings.
        struct_forwards = common + [
            "codec",
            "stream",
            "methods",
            "default_endianness",
        ]
        struct: dict[str, Any]
        for struct in data.get("structs", {}).values():  # type: ignore
            for forward in struct_forwards:
                struct.setdefault(
                    forward,
                    data["struct"][forward],  # type: ignore
                )


def load(path: Pathlike) -> Config:
    """Load a configuration object."""

    src_config = find_file("default.yaml", package=PKG_NAME)
    assert src_config is not None

    data = merge(
        _ARBITER.decode(
            src_config,
            includes_key=DEFAULT_INCLUDES_KEY,
            require_success=True,
        ).data,
        _ARBITER.decode(
            path, includes_key=DEFAULT_INCLUDES_KEY, require_success=True
        ).data,
        # Always allow the project-specific configuration to override package
        # data.
        expect_overwrite=True,
    )

    return Config.create(data)



"""
A module for generating enumeration header files.
"""

# built-in
from typing import Optional

# third-party
from vcorelib.io import IndentedFileWriter

# internal
from ifgen.enum.common import enum_to_string_function, string_to_enum_function
from ifgen.enum.python import python_enum_header
from ifgen.generation.interface import GenerateTask
from ifgen.generation.json import to_json_method

EnumConfig = Optional[dict[str, int | str]]


def enum_line(name: str, value: EnumConfig) -> str:
    """Build a string representing a line in an enumeration."""

    line = name

    if value and value.get("value"):
        line += f" = {value['value']}"

    if value and value.get("description"):
        line += f" /*!< {value['description']} */"

    return line


def cpp_enum_neader(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Create a C++ enumeration header."""

    writer.write(f"enum class {task.name} : {task.instance['underlying']}")
    with writer.scope(suffix=";"):
        writer.join(
            *(
                enum_line(enum, value)
                for enum, value in task.instance.get("enum", {}).items()
            )
        )

    writer.write(
        (
            f"static_assert(sizeof({task.name}) == "
            f"{task.env.size(task.name)});"
        )
    )

    runtime = task.enum()

    writer.empty()

    if task.instance["identifier"]:
        writer.write(
            f"static constexpr enum_id_t {task.name}_id = {runtime.id};"
        )
        writer.empty()

    if runtime.default:
        writer.write(
            f"static constexpr auto {task.name}_default = "
            f"{task.name}::{task.instance['default']};"
        )
        writer.empty()

    enum_to_string_function(
        task, writer, task.instance["use_map"], definition=True
    )

    with writer.padding():
        writer.write(
            "inline std::ostream &operator<<(std::ostream &stream, "
            f"{task.name} instance)"
        )
        with writer.scope():
            writer.write("stream << to_string(instance);")
            writer.write("return stream;")

    string_to_enum_function(
        task, writer, task.instance["use_map"], definition=True
    )

    if task.instance["json"]:
        to_json_method(
            task,
            writer,
            runtime.asdict(),
            dumps_indent=task.instance["json_indent"],
            definition=True,
            inline=True,
        )


def enum_header(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Create a header file for an enumeration."""

    if task.is_cpp:
        cpp_enum_neader(task, writer)
    elif task.is_python:
        python_enum_header(task, writer)



"""
A module exposing some pathing utilities.
"""

# built-in
from pathlib import Path
import shutil
import subprocess

# third-party
from vcorelib.paths import Pathlike, normalize
from vcorelib.paths.context import TextPreprocessor


def combine_if_not_absolute(root: Path, candidate: Pathlike) -> Path:
    """Combine a root directory with a path if the path isn't absolute."""

    candidate = normalize(candidate)
    return candidate if candidate.is_absolute() else root.joinpath(candidate)


def audit_init_file(source: Path, parent_depth: int = 0) -> None:
    """Create initialization files if necessary."""

    candidate = source.parent.joinpath("__init__.py")
    if not candidate.exists():
        with candidate.open("wb") as stream:
            stream.write(bytes())

    # Audit parent directories.
    if parent_depth > 0:
        parent_depth -= 1
        audit_init_file(source.parent, parent_depth=parent_depth)


def create_formatter(*args: str, **kwargs) -> TextPreprocessor:
    """
    Create a formatting preprocessor that uses stdin/stdout of a subprocess.
    """

    def formatter(data: str) -> str:
        """Run clang-format on the provided data."""

        result = data

        if args and shutil.which(args[0]) is not None:
            result = subprocess.run(
                args,
                input=data,
                text=True,
                capture_output=True,
                check=True,
                **kwargs,
            ).stdout

        return result

    return formatter



"""
A struct-receiver interface implementation.
"""

# built-in
from typing import Any

# third-party
from vcorelib.io.arbiter import ARBITER
from vcorelib.names import to_snake
from vcorelib.paths import rel

# internal
from ifgen.enums import Language
from ifgen.generation.interface import GenerateTask
from ifgen.generation.python import python_imports
from ifgen.struct import header_for_type
from ifgen.struct.python import python_dependencies


def python_struct_receiver(task: GenerateTask) -> None:
    """Python struct receiver generation."""

    structs: list[str] = list(task.env.config.data.get("structs", {}))

    with task.boilerplate(json=False, parent_depth=2) as writer:
        python_imports(
            writer,
            third_party={
                "runtimepy.codec.protocol.receiver": ["StructReceiver"]
            },
            internal=python_dependencies(
                [], structs, struct_prefix="..structs"
            ),
            final_empty=1,
        )

        writer.write("RECEIVER = StructReceiver(")
        with writer.indented():
            for struct in structs:
                writer.write(struct + ",")
        writer.write(")")

    runtimepy_structs: list[Any] = []
    for struct in structs:
        snake = to_snake(struct)
        runtimepy_structs.append(
            {
                "name": snake,
                "config": {
                    "protocol_factory": ".".join(
                        list(
                            rel(
                                task.env.directories[Language.PYTHON].output,
                                base=task.env.root_path,
                            ).parts
                        )
                        + ["structs", snake, struct]
                    )
                },
            }
        )

    assert ARBITER.encode(
        task.path.with_suffix(".yaml"), {"structs": runtimepy_structs}
    )[0]


# pylint: disable=too-many-statements
def cpp_struct_receiver(task: GenerateTask) -> None:
    """C++ struct receiver generation."""

    data = task.env.config.data
    structs: list[str] = list(
        x for x in data.get("structs", {}) if data["structs"][x]["codec"]
    )

    names_qualified = {}
    for struct in structs:
        names = data["structs"][struct].get("namespace", [])
        names.append(struct)
        names_qualified[struct] = "::".join(names)

    with task.boilerplate(
        json=False,
        includes=["<functional>"]
        + [header_for_type(x, task) for x in structs],
        parent_depth=2,
    ) as writer:
        writer.write("template <ifgen_struct T>")
        writer.write("using struct_handler = std::function<void(const T &)>;")
        writer.empty()

        writer.write("using non_struct_handler =")
        with writer.indented():
            writer.write(
                "std::function<std::size_t(const std::byte *, std::size_t)>;"
            )
        writer.empty()

        writer.write("struct StructReceiver")
        with writer.scope(suffix=";"):
            for struct in structs:
                snake = to_snake(struct)
                writer.write(f"{names_qualified[struct]} {snake};")
                writer.write(
                    f"struct_handler<decltype({snake})> "
                    f"{snake}_handler = nullptr;"
                )
                writer.empty()

            writer.write("non_struct_handler non_struct = nullptr;")
            writer.empty()

            writer.write("std::size_t dropped_messages = 0;")
            writer.write("std::size_t dropped_bytes = 0;")
            writer.empty()

            writer.write("inline void drop_message(std::size_t &len)")
            with writer.scope():
                writer.write("if (len)")
                with writer.scope():
                    writer.write("dropped_messages++;")
                    writer.write("dropped_bytes += len;")
                    writer.write("len = 0;")

            writer.empty()
            writer.write("template <std::endian endianness = default_endian>")
            writer.write(
                "void handle_message(const std::byte *data, std::size_t len)"
            )
            with writer.scope():
                writer.write("if (len < sizeof(struct_id_t))")
                with writer.scope():
                    writer.write("drop_message(len);")
                    writer.write("return;")

                writer.empty()

                writer.c_comment("Read identifier and advance buffer.")
                writer.write("struct_id_t ident = handle_endian<endianness>(")
                with writer.indented():
                    writer.write(
                        "*reinterpret_cast<const struct_id_t *>(data));"
                    )
                writer.write("data += sizeof(struct_id_t);")
                writer.write("len -= sizeof(struct_id_t);")

                writer.empty()

                writer.write("switch (ident)")
                writer.write("{")

                writer.write("case 0:")
                with writer.indented():
                    writer.write("if (non_struct)")
                    with writer.scope():
                        writer.write("auto result = non_struct(data, len);")
                        writer.write("if (result)")
                        with writer.scope():
                            writer.write("data += result;")
                            writer.write("len -= result;")
                        writer.write("else")
                        with writer.scope():
                            writer.write("drop_message(len);")

                    writer.write("else")
                    with writer.scope():
                        writer.write("drop_message(len);")

                    writer.write("break;")

                for struct in structs:
                    snake = to_snake(struct)
                    writer.write(f"case decltype({snake})::id:")
                    with writer.indented():
                        writer.write(f"if (len >= decltype({snake})::size)")
                        with writer.scope():
                            writer.write(f"if ({snake}_handler)")
                            with writer.scope():
                                writer.write(f"{snake}.decode<endianness>(")
                                with writer.indented():
                                    writer.write(
                                        f"reinterpret_cast<const decltype"
                                        f"({snake}) *>(data)->raw_ro());"
                                    )
                                writer.write(f"{snake}_handler({snake});")

                            writer.write(f"data += decltype({snake})::size;")
                            writer.write(f"len -= decltype({snake})::size;")

                        writer.write("else")
                        with writer.scope():
                            writer.write("drop_message(len);")

                        writer.write("break;")

                writer.write("default:")
                with writer.indented():
                    writer.c_comment("Couldn't match any identifier.")
                    writer.write("drop_message(len);")
                writer.write("}")

                writer.empty()

                writer.c_comment("Continue if more bytes remain.")
                writer.write("if (len)")
                with writer.scope():
                    writer.write("handle_message<endianness>(data, len);")
